package proxy_test

import (
	"net/http"
	"net/http/httptest"
	"sync/atomic"
	"time"

	"code.cloudfoundry.org/loggregator-release/src/trafficcontroller/internal/proxy"

	"github.com/gorilla/websocket"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

var _ = Describe("KeepAlive", func() {
	var (
		testServer         *httptest.Server
		wsClient           *websocket.Conn
		keepAliveCompleted chan struct{}
	)

	BeforeEach(func() {
		keepAliveCompleted = make(chan struct{})
		testServer = httptest.NewServer(makeTestHandler(keepAliveCompleted))
		var err error
		wsClient, _, err = websocket.DefaultDialer.Dial(httpToWs(testServer.URL), nil)
		Expect(err).NotTo(HaveOccurred())
	})

	AfterEach(func() {
		wsClient.Close()
		testServer.Close()
	})

	It("sends pings to the client", func() {
		var pingCount int32
		wsClient.SetPingHandler(func(string) error {
			atomic.AddInt32(&pingCount, 1)
			return nil
		})

		go func() {
			defer GinkgoRecover()
			_, _, err := wsClient.ReadMessage()
			Expect(err.Error()).To(ContainSubstring("closed network connection"))
		}()
		Eventually(func() int32 { return atomic.LoadInt32(&pingCount) }).ShouldNot(BeZero())
	})

	It("doesn't close the client when it responds with pong frames", func() {
		go func() {
			defer GinkgoRecover()
			_, _, err := wsClient.ReadMessage()
			Expect(err.Error()).To(ContainSubstring("closed network connection"))
		}()
		// default ping handler responds with pong frames
		Consistently(keepAliveCompleted).ShouldNot(BeClosed())
	})

	It("closes the client when it doesn't respond with pong frames", func() {
		wsClient.SetPingHandler(func(string) error { return nil })
		go func() {
			defer GinkgoRecover()
			_, _, err := wsClient.ReadMessage()
			Expect(err.Error()).To(ContainSubstring("closed network connection"))
		}()
		Eventually(keepAliveCompleted).Should(BeClosed())
	})
})

func makeTestHandler(keepAliveCompleted chan struct{}) http.Handler {
	return http.HandlerFunc(func(rw http.ResponseWriter, req *http.Request) {
		conn, _ := (&websocket.Upgrader{}).Upgrade(rw, req, nil)
		go func() {
			defer GinkgoRecover()
			_, _, _ = conn.ReadMessage()
		}()
		proxy.NewKeepAlive(conn, 50*time.Millisecond).Run()
		close(keepAliveCompleted)
	})
}
